#include "matrix1.h"
#include <iostream>
#include <algorithm>
#include <initializer_list>
#include <cmath>
#include <vector>
#include <iomanip>

namespace linalg {
    // Конструкторы
    Matrix::Matrix() : m_ptr(nullptr), m_rows(0), m_columns(0) {}

    Matrix::Matrix(int rows) : m_rows(rows), m_columns(1) {
        if (rows <= 0) {
            throw std::runtime_error("Строки и столбцы должны быть положительными!");
        }
        m_ptr = new double[rows];
    }

    Matrix::Matrix(int rows, int columns) : m_rows(rows), m_columns(columns) {
        if (rows <= 0 || columns <= 0) {
            throw std::runtime_error("Строки и столбцы должны быть положительными!");
        }
        m_ptr = new double[rows * columns];
    }

    Matrix::Matrix(const Matrix& other) : m_rows(other.m_rows), m_columns(other.m_columns) {
        m_ptr = new double[m_rows * m_columns];
        std::copy(other.m_ptr, other.m_ptr + (m_rows * m_columns), m_ptr);
    }

    Matrix::Matrix(Matrix&& other) noexcept : m_ptr(other.m_ptr), m_rows(other.m_rows), m_columns(other.m_columns) {
        other.m_ptr = nullptr;
        other.m_rows = 0;
        other.m_columns = 0;
    }

    Matrix::Matrix(std::initializer_list<std::initializer_list<double>> init_list) {
        m_rows = init_list.size();
        m_columns = init_list.begin()->size();

        for (const auto& row : init_list) {
            if (row.size() != static_cast<size_t>(m_columns)) {
                throw std::invalid_argument("Непостоянное количество столбцов в init_list.");
            }
        }

        m_ptr = new double[m_rows * m_columns];
        int i = 0;
        for (const auto& row : init_list) {
            std::copy(row.begin(), row.end(), m_ptr + i * m_columns);
            i++;
        }
    }

    // Деструктор
    Matrix::~Matrix() {
        delete[] m_ptr;
    }


// Метод для получения количества строк
    int Matrix::rows() const {
        return m_rows;
    }

    // Метод для получения количества колонок
    int Matrix::columns() const {
        return m_columns;
    }

    // Метод для проверки, является ли матрицей пустой
    bool Matrix::empty() const {
        return m_rows == 0 || m_columns == 0;
    }

    // Метод для изменения размерности матрицы
    void Matrix::reshape(int rows, int columns) {
        try {
            if (rows * columns != m_rows * m_columns || rows <= 0 || columns <= 0) {
                throw std::runtime_error("Новые строки и столбцы должны быть положительными и новая матрица должна иметь такое же количество элементов.");
            } else {
                m_rows = rows;
                m_columns = columns;
            }
        } catch (const std::runtime_error& e) {
            std::cerr << "Error: " << e.what() << std::endl;
        }
    }

    std::ostream& operator<<(std::ostream& os, const Matrix& matrix) {
        std::vector<int> maxIntPartWidth(matrix.m_columns, 0);

        // Найдем максимальную ширину для целой и дробной части каждого столбца
        for (int i = 0; i < matrix.m_rows; i++) {
            for (int j = 0; j < matrix.m_columns; j++) {
                double value = matrix.m_ptr[i * matrix.m_columns + j];
                int intPartWidth = static_cast<int>(std::log10(std::abs(value))) + 1;
                if (intPartWidth > maxIntPartWidth[j]) {
                    maxIntPartWidth[j] = intPartWidth;
                }
            }
        }

        // Выводим матрицу с соответствующим выравниванием
        for (int i = 0; i < matrix.m_rows; i++) {
            os << "| ";
            for (int j = 0; j < matrix.m_columns; j++) {
                os << std::setw(maxIntPartWidth[j] + 4) << std::fixed << std::setprecision(3) << matrix.m_ptr[i * matrix.m_columns + j] << ' ';
            }
            os << "|\n";
        }
        return os;
    }

    Matrix& Matrix::operator=(const Matrix& other) {
        if (this != &other) {
            delete[] m_ptr; // Освобождаем ресурсы текущей матрицы

            m_rows = other.m_rows;
            m_columns = other.m_columns;
            m_ptr = new double[m_rows * m_columns];

            std::copy(other.m_ptr, other.m_ptr + m_rows * m_columns, m_ptr);
        }
        return *this;
    }

    // Оператор присваивания (перемещающий)
    Matrix& Matrix::operator=(Matrix&& other) noexcept {
        if (this != &other) {
            delete[] m_ptr; // Освобождаем ресурсы текущей матрицы

            m_rows = other.m_rows;
            m_columns = other.m_columns;
            m_ptr = other.m_ptr;

            other.m_rows = 0;
            other.m_columns = 0;
            other.m_ptr = nullptr;
        }
        return *this;
    }

    double& Matrix::operator()(int row, int col) {
        try {
            if (row < 0 || row >= m_rows || col < 0 || col >= m_columns) {
                throw std::runtime_error("Индекс неправильный!");
            } else {
                return m_ptr[row * m_columns + col];
            }
        } catch (const std::runtime_error& e) {
            std::cerr << "Error: " << e.what() << std::endl;
            return m_ptr[0];
        }
    }

    const double& Matrix::operator()(int row, int col) const {
        try {
            if (row < 0 || row >= m_rows || col < 0 || col >= m_columns) {
                throw std::runtime_error("Индекс неправильный!");
            } else {
                return m_ptr[row * m_columns + col];
            }
        } catch (const std::runtime_error& e) {
            std::cerr << "Error: " << e.what() << std::endl;
            return m_ptr[0];
        }
    }

    Matrix Matrix::operator+(const Matrix &other) const {

        try {
            if (m_rows != other.m_rows || m_columns != other.m_columns) {
                throw std::runtime_error("Матрицы с таким размероа нельзя складывать или вычитать.");
            } else {
                Matrix result(m_rows, m_columns);
                for (int i = 0; i < m_rows; i++) {
                    for (int j = 0; j < m_columns; j++) {
                        result(i, j) = m_ptr[i * m_columns + j] + other(i, j);
                    }
                }
                return result;
            }
        } catch (const std::runtime_error& e) {
            std::cerr << "Error: " << e.what() << std::endl;
            return other;
        }
    }

    // Перегрузка оператора +=
    Matrix &Matrix::operator+=(const Matrix &other) {
        try {
            if (m_rows != other.m_rows || m_columns != other.m_columns) {
                throw std::runtime_error("Матрицы с таким размероа нельзя складывать или вычитать.");
            } else {
                for (int i = 0; i < m_rows; i++) {
                    for (int j = 0; j < m_columns; j++) {
                        m_ptr[i * m_columns + j] += other(i, j);
                    }
                }
                return *this;
            }
        } catch (const std::runtime_error& e) {
            std::cerr << "Error: " << e.what() << std::endl;
            return *this;
        }
    }

    // Перегрузка оператора -
    Matrix Matrix::operator-(const Matrix &other) const {
        try {
            if (m_rows != other.m_rows || m_columns != other.m_columns) {
                throw std::runtime_error("Матрицы с таким размероа нельзя складывать или вычитать.");
            } else {
                Matrix result(m_rows, m_columns);
                for (int i = 0; i < m_rows; i++) {
                    for (int j = 0; j < m_columns; j++) {
                        result(i, j) = m_ptr[i * m_columns + j] - other(i, j);
                    }
                }
                return result;
            }
        } catch (const std::runtime_error& e) {
            std::cerr << "Error: " << e.what() << std::endl;
            return other;
        }
    }

    // Перегрузка оператора -=
    Matrix &Matrix::operator-=(const Matrix &other) {
        try {
            if (m_rows != other.m_rows || m_columns != other.m_columns) {
                throw std::runtime_error("Матрицы с таким размероа нельзя складывать или вычитать.");
            } else {
                for (int i = 0; i < m_rows; i++) {
                    for (int j = 0; j < m_columns; j++) {
                        m_ptr[i * m_columns + j] -= other(i, j);
                    }
                }
                return *this;
            }
        } catch (const std::runtime_error& e) {
            std::cerr << "Error: " << e.what() << std::endl;
            return *this;
        }
    }

    // Перегрузка оператора *
    Matrix Matrix::operator*(const Matrix &other) const {
        try {
            if (m_rows != other.m_rows || m_columns != other.m_columns) {
                throw std::runtime_error("Матрицы с таким размероа нельзя перемножать.");
            } else {
                Matrix result(m_rows, other.m_columns);
                for (int i = 0; i < m_rows; i++) {
                    for (int j = 0; j < other.m_columns; j++) {
                        double sum = 0.0;
                        for (int k = 0; k < m_columns; k++) {
                            sum += m_ptr[i * m_columns + k] * other(k, j);
                        }
                        result(i, j) = sum;
                    }
                }
                return result;
            }
        } catch (const std::runtime_error& e) {
            std::cerr << "Error: " << e.what() << std::endl;
            return other;
        }
    }

    // Перегрузка оператора * для умножения на число
    Matrix Matrix::operator*(double scalar) const {
        Matrix result(m_rows, m_columns);
        for (int i = 0; i < m_rows; i++) {
            for (int j = 0; j < m_columns; j++) {
                result(i, j) = m_ptr[i * m_columns + j] * scalar;
            }
        }
        return result;
    }

    // Перегрузка оператора * для умножения на число (справа)
    Matrix operator*(double scalar, const Matrix &matrix) {
        return matrix * scalar;
    }

    // Перегрузка оператора *=
    Matrix &Matrix::operator*=(const Matrix &other) {
        try {
            if (m_rows != other.m_rows || m_columns != other.m_columns) {
                throw std::runtime_error("Матрицы с таким размероа нельзя перемножать.");
            } else {
                *this = *this * other;
                return *this;
            }
        } catch (const std::runtime_error& e) {
            std::cerr << "Error: " << e.what() << std::endl;
            return *this;
        }
    }

    // Перегрузка оператора *= для умножения на число
    Matrix &Matrix::operator*=(double scalar) {
        for (int i = 0; i < m_rows; i++) {
            for (int j = 0; j < m_columns; j++) {
                m_ptr[i * m_columns + j] *= scalar;
            }
        }
        return *this;
    }

    // Перегрузка оператора ==
    bool Matrix::operator==(const Matrix &other) const {
        if (m_rows != other.m_rows || m_columns != other.m_columns) {
            return false;
        }
        for (int i = 0; i < m_rows; i++) {
            for (int j = 0; j < m_columns; j++) {
                if (m_ptr[i * m_columns + j] != other(i, j)) {
                    return false;
                }
            }
        }
        return true;
    }

    // Перегрузка оператора !=
    bool Matrix::operator!=(const Matrix &other) const {
        return !(*this == other);
    }
    double Matrix::norm() const {
        if (empty()) {
            return 0.0;
        }

        double sum = 0.0;
        for (int i = 0; i < m_rows; ++i) {
            for (int j = 0; j < m_columns; ++j) {
                sum += m_ptr[i * m_columns + j] * m_ptr[i * m_columns + j];
            }
        }
        return std::sqrt(sum);
    }

    double Matrix::trace() const {
        double sum = 0.0;
        for (int i = 0; i < m_rows; ++i) {
            sum += m_ptr[i * m_columns + i];
        }
        return sum;
    }

}